import torch
import numpy as np
from torch import Tensor
from typing import Tuple, Union

from ._core import TError


class Memory(object):
    """A container for node memory."""

    def __init__(self, num_nodes: int, dim: int, device=None):
        """
        Internal constructor for creating a Memory. Initialize node memory and timestamps to zero.

        :param int num_nodes: Number of nodes
        :param int dim: Length of memory vector for a single node
        :param device: Which device to put node memory
        :type device: None or str or torch.device
        """
        self._device = torch.device('cpu' if device is None else device)

        self._data = torch.zeros((num_nodes, dim), device=self._device)
        self._time = torch.zeros(num_nodes, device=self._device)

        if list(self._data.shape) != [num_nodes, dim]:
            raise TError('memory data dimension mismatch')
        if self._time.shape[0] != num_nodes:
            raise TError('memory timestamp dimension mismatch')

    def __len__(self) -> int:
        """
        Return number of nodes in Memory

        :rtype: int
        """
        return self._data.shape[0]

    @property
    def data(self) -> Tensor:
        """Return node memory

        :rtype: Tensor
        """
        return self._data

    @property
    def time(self) -> Tensor:
        """Return timestamps of current node memory

        :rtype: Tensor
        """
        return self._time

    @property
    def device(self) -> torch.device:
        """Return the device where Memory is located

        :rtype: torch.device
        """
        return self._device

    def dim(self) -> int:
        """Return length of memory vector for a single node"""
        return self._data.shape[1]

    def reset(self):
        """Reset node memory and timestamps to zero"""
        self._data.zero_()
        self._time.zero_()

    def update(self, nids: Union[np.ndarray, Tensor], newdata: Tensor, newtime: Tensor):
        if not isinstance(nids, Tensor):
            nids = torch.from_numpy(nids).long()
        nids = nids.to(self._device)
        self._data[nids] = newdata.detach().to(self._device)
        self._time[nids] = newtime.detach().to(self._device)

    def move_to(self, device, **kwargs):
        if device is None or self._device == device:
            return
        self._data = self._data.to(device, **kwargs)
        self._time = self._time.to(device, **kwargs)
        self._device = device

    def backup(self) -> Tuple[Tensor, Tensor]:
        return (self._data.cpu().clone(), self._time.cpu().clone())

    def restore(self, state: Tuple[Tensor, Tensor]):
        data, time = state
        if self._data.shape != data.shape:
            raise TError('memory data dimension mismatch')
        if self._time.shape != time.shape:
            raise TError('memory timestamp dimension mismatch')
        self._data = data.clone().to(self._device)
        self._time = time.clone().to(self._device)



from __future__ import annotations
from typing import TYPE_CHECKING, Callable, Optional, Tuple
if TYPE_CHECKING:
    from ._graph import TGraph
    from ._context import TContext

import torch
import numpy as np
from torch import Tensor

from ._core import TError
from ._frame import TFrame
from ._stats import tt


class TBlock(object):
    """Captures 1-hop relations between node/time pairs and their neighbors for doing computations, such as segmented
     softmax and message-passing aggregation."""

    def __init__(self, ctx: 'TContext', layer: int, dstnodes: np.ndarray, dsttimes: np.ndarray,
                 dstindex: np.ndarray = None, srcnodes: np.ndarray = None,
                 eid: np.ndarray = None, ets: np.ndarray = None):
        """
        Internal constructor for creating a TBlock.

        :param TContext ctx: The TContext.
        :param int layer: The layer in the GNN.
        :param np.ndarray dstnodes: The destination nodes.
        :param np.ndarray dstindex: The indices of destination nodes in dstnodes.
        :param np.ndarray srcnodes: The source nodes of the edges.
        :param np.ndarray eid: The edge indices.
        :param np.ndarray ets: The edge timestamps.
        """

        self._ctx = ctx
        self._g = ctx.graph()
        self._layer = layer
        self._dstnodes = dstnodes
        self._dsttimes = dsttimes

        # metadata
        self._prev = None
        self._next = None
        self._hooks = []
        self._include_prev_dst = False

        # neighbor attributes
        self._dstindex = dstindex
        self._srcnodes = srcnodes
        self._eid = eid
        self._ets = ets
        self._has_nbrs = (
            self._dstindex is not None and
            self._srcnodes is not None and
            self._eid is not None and
            self._ets is not None)

        # data attributes
        self._dstdata = TFrame(len(dstnodes))
        self._srcdata = TFrame(0 if srcnodes is None else len(srcnodes))
        self._edata = TFrame(0 if eid is None else len(eid))

        # cached attributes
        self._c_efeat = None
        self._c_nfeat = None
        self._c_allnodes = None
        self._c_uniq_src = None
        self._c_mem_data = None
        self._c_mail = None

    @property
    def g(self) -> 'TGraph':
        """Returns the TGraph associated with this TBlock."""
        return self._g

    @property
    def layer(self) -> int:
        """Returns the GNN layer to which this TBlock belongs."""
        return self._layer
    
    @property
    def dstnodes(self) -> np.ndarray:
        """Returns the full destination nodes."""
        return self._dstnodes

    @property
    def dsttimes(self) -> np.ndarray:
        """Returns the destination node timestamps."""
        return self._dsttimes

    @property
    def dstindex(self) -> Optional[np.ndarray]:
        """Returns the indices of destination nodes in self.dstnodes() that are used for sampling."""
        return self._dstindex

    @property
    def srcnodes(self) -> Optional[np.ndarray]:
        """Return the source nodes (temporally sampled neighbors)."""
        return self._srcnodes

    @property
    def eid(self) -> Optional[np.ndarray]:
        """Returns the edge indices, which are connections towards sampled neighbors."""
        return self._eid

    @property
    def ets(self) -> Optional[np.ndarray]:
        """Returns the edge timestamps."""
        return self._ets
    
    @property
    def dstdata(self) -> TFrame:
        """Returns the destination node data."""
        return self._dstdata

    @property
    def srcdata(self) -> TFrame:
        """Returns the source node data."""
        return self._srcdata
    
    @property
    def edata(self) -> TFrame:
        """Returns the edge data."""
        return self._edata

    @property
    def prev(self) -> Optional[TBlock]:
        """Returns the previous TBlock."""
        return self._prev

    @prev.setter
    def prev(self, value: 'TBlock'):
        """
        Sets the previous TBlock.

        :param TBlock value: The previous TBlock.
        :raises TError: If value is not an instance of TBlock or its layer is not the current TBlock's layer - 1.
        """
        if not isinstance(value, TBlock) or value.layer != self.layer - 1:
            raise TError('invalid previous block')
        self._prev = value

    @property
    def next(self) -> Optional[TBlock]:
        """Returns the next TBlock."""
        return self._next

    @next.setter
    def next(self, value: 'TBlock'):
        """
        Sets the next TBlock.

        :param TBlock value: The next TBlock.
        :raises TError: If value is not an instance of TBlock or its layer is not current TBlock's layer + 1.
        """
        if not isinstance(value, TBlock) or value.layer != self.layer + 1:
            raise TError('invalid next block')
        self._next = value

    def num_edges(self) -> int:
        """Returns the number of edges in the block when self.has_nbrs() is True."""
        return len(self._eid) if self._has_nbrs else 0

    def num_src(self) -> int:
        """Returns the length of source node array when self.has_nbrs() is True."""
        return len(self._srcnodes) if self._has_nbrs else 0

    def num_dst(self) -> int:
        """Returns the length of destination node array."""
        return len(self._dstnodes)

    def has_nbrs(self) -> bool:
        """Returns True when the block has the neighbor attributes, including dstindex, srcnodes, eid and ets."""
        return self._has_nbrs

    def set_nbrs(self, dstindex: np.ndarray, srcnodes: np.ndarray,
                      eid: np.ndarray, ets: np.ndarray):
        """Sets the neighbor attributes for the block."""
        self.clear_nbrs()
        self._has_nbrs = True
        self._dstindex = dstindex
        self._srcnodes = srcnodes
        self._eid = eid
        self._ets = ets
        self._edata = TFrame(len(eid))
        self._srcdata = TFrame(len(srcnodes))

    def clear_nbrs(self):
        """Clears the neighbor attributes and related cache."""
        self._has_nbrs = False
        self._dstindex = None
        self._srcnodes = None
        self._eid = None
        self._ets = None
        self._edata = TFrame(0)
        self._srcdata = TFrame(0)
        self._c_efeat = None
        self._c_nfeat = None
        self._c_allnodes = None
        self._c_uniq_src = None

    def clear_data(self):
        """Clears the data and all cache."""
        self._edata.clear()
        self._srcdata.clear()
        self._dstdata.clear()
        self._c_efeat = None
        self._c_nfeat = None
        self._c_allnodes = None
        self._c_uniq_src = None
        self._c_mem_data = None
        self._c_mail = None

    def next_block(self, include_dst=False, use_dst_times=False) -> TBlock:
        """
        Generates the next TBlock with neighbor information.

        :param bool include_dst: Whether to include current TBlocks's destination nodes in the next TBlock.
        :param bool use_dst_times: True when using timestamps of current TBlock's destination nodes as timestamps of
         next TBlock's destination nodes, False when using edges' timestamps.
        """
        if self._next is not None:
            return self._next
        t_start = tt.start()
        if self.num_dst() == 0:
            blk = TBlock(self._ctx, self._layer + 1, np.empty(0), np.empty(0))
        else:
            self._check_has_nbrs()
            next_dstnodes = self._srcnodes
            next_dsttimes = self._ets if not use_dst_times else self._dsttimes[self._dstindex]
            if include_dst:
                next_dstnodes = np.concatenate([self._dstnodes, next_dstnodes])
                next_dsttimes = np.concatenate([self._dsttimes, next_dsttimes])
            blk = TBlock(self._ctx, self._layer + 1, next_dstnodes, next_dsttimes)
        blk._include_prev_dst = True
        blk._prev = self
        self._next = blk
        tt.t_prep_input += tt.elapsed(t_start)
        return blk

    def allnodes(self) -> Tensor:
        """Returns a tensor containing the destination nodes concatenated with the source nodes (if available) in pre-defined TGraph's storage device."""
        if self._c_allnodes is None:
            sdev = self._g.storage_device()
            nodes = np.concatenate([self._dstnodes, self._srcnodes]) \
                if self._has_nbrs else self._dstnodes
            nodes = torch.from_numpy(nodes).long().to(sdev)
            self._c_allnodes = nodes
        return self._c_allnodes

    def uniq_src(self) -> Optional[Tuple[Tensor, Tensor]]:
        """Returns a Tensor tuple: (the unique source node indices, the indices of the unique array that reconstructs the source
         node array) in pre-defined TGraph's storage device."""
        if self._c_uniq_src is None and self._srcnodes is not None:
            sdev = self._g.storage_device()
            uniq_nids, idx = np.unique(self._srcnodes, return_inverse=True)
            uniq_nids = torch.from_numpy(uniq_nids).long().to(sdev)
            idx = torch.from_numpy(idx).long().to(sdev)
            self._c_uniq_src = (uniq_nids, idx)
        return self._c_uniq_src

    def efeat(self) -> Optional[Tensor]:
        """Returns the edge features in TGraph's computation device, always use pinned memory if possible."""
        self._load_efeat(use_pin=True)
        return self._c_efeat

    def nfeat(self) -> Optional[Tensor]:
        """Returns the node features in TGraph's computation device, always use pinned memory if possible."""
        self._load_nfeat(use_pin=True)
        return self._c_nfeat

    def srcfeat(self) -> Optional[Tensor]:
        """Returns the source node features in TGraph's computation device, always use pinned memory if possible."""
        self._load_nfeat(use_pin=True)
        return self._c_nfeat[self.num_dst():]

    def dstfeat(self) -> Optional[Tensor]:
        """Returns the destination node features in TGraph's computation device, always use pinned memory if possible."""
        self._load_nfeat(use_pin=True)
        return self._c_nfeat[:self.num_dst()]

    def mem_data(self) -> Optional[Tensor]:
        """Returns node memory in TGraph's computation device, always use pinned memory if possible."""
        self._load_mem_data(use_pin=True)
        return self._c_mem_data

    def mail(self) -> Optional[Tensor]:
        """Returns the node mails in TGraph's computation device, always use pinned memory if possible."""
        self._load_mail(use_pin=True)
        return self._c_mail

    def time_deltas(self) -> Tensor:
        """Computes the timestamp differences between destination nodes (used for sampling) and edges and returns
        them as a tensor in pre-defined TGraph's computation device.
        
        :return: A tensor containing the time deltas.
        """
        self._check_has_nbrs()
        dts = self._dsttimes[self._dstindex]
        dts = dts - self._ets
        dev = self._g.compute_device()
        return torch.from_numpy(dts).to(device=dev, dtype=torch.float)

    def apply(self, fn: Callable, need_nbrs=True, run_hooks=True):
        """
        Applies an operator to the block itself and returns the output.

        :param Callable fn: The function to apply to the block itself.
        :param bool need_nbrs: Whether sampled neighbors are needed.
        :param run_hooks: Whether to run registered hooks.
        """
        if need_nbrs:
            self._check_has_nbrs()
        output = fn(self)
        if run_hooks:
            output = self.run_hooks(output)
        return output

    def run_hooks(self, input: Optional[Tensor]) -> Optional[Tensor]:
        """Runs registered hooks on the input tensor in reversed order."""
        for hook in reversed(self._hooks):
            input = hook(self, input)
        return input

    def clear_hooks(self):
        """Clears all the hooks."""
        self._hooks.clear()

    def register_hook(self, hook: Callable):
        """Registers a hook for running post-processing."""
        self._hooks.append(hook)

    def _check_has_nbrs(self):
        """
        Makes sure the block has all the neighbor attributes, including dstindex, srcnodes, eid and ets.

        :raises TError: If self.has_neighbor is False.
        """
        if not self._has_nbrs:
            raise TError('block neighbors has not been sampled yet')

    def _replace_dst_empty(self):
        """Replaces the destination nodes and timestamps with empty arrays."""
        self._replace_dst(np.empty(0), np.empty(0))

    def _replace_dst(self, dstnodes: np.ndarray, dsttimes: np.ndarray):
        """Replaces destination nodes and timestamps with given arrays."""
        self.clear_nbrs()
        self.clear_data()
        self._dstnodes = dstnodes
        self._dsttimes = dsttimes
        self._dstdata = TFrame(len(dstnodes))

    def _load_efeat(self, use_pin=False):
        """Loads the edge features to the TGraph's computation device."""
        if self._c_efeat is None and self._g.efeat is not None:
            self._c_efeat = self._load_feat(
                self._g.efeat, self._eid, use_pin,
                self._ctx._get_efeat_pin)

    def _load_nfeat(self, use_pin=False):
        """Loads the node features to the TGraph's computation device."""
        if self._c_nfeat is None and self._g.nfeat is not None:
            self._c_nfeat = self._load_feat(
                self._g.nfeat, self.allnodes().cpu().numpy(), use_pin,
                self._ctx._get_nfeat_pin)

    def _load_feat(self, feat: Tensor, idx: np.ndarray, use_pin: bool, pin_getter: Callable) -> Tensor:
        """Loads selected feature data from the TGraph's storage device to computation device.

        :param Tensor feat: Feature tensor.
        :param np.ndarray idx: The indices of selected features.
        :param bool use_pin: Whether to use pinned memory, only applicable when the storage device is cpu and the computation device.
        is cuda.
        :param Callable pin_getter: Function to get the pinned buffer.
        :return: A tensor containing the loaded feature data.
        """
        t_start = tt.start()
        sdev = self._g.storage_device()
        cdev = self._g.compute_device()
        idx = torch.from_numpy(idx).long()
        if sdev.type == 'cpu' and cdev.type == 'cuda' and use_pin:
            pin = pin_getter(self.layer, len(idx), feat.shape[1])
            torch.index_select(feat, 0, idx, out=pin)
            data = pin.to(cdev, non_blocking=True)
        else:
            data = feat[idx.to(sdev)].to(cdev)
        tt.t_prep_input += tt.elapsed(t_start)
        return data

    def _load_mem_data(self, use_pin=False):
        """Loads the node memory to the TGraph's computation device."""
        if self._c_mem_data is None and self._g.mem is not None:
            t_start = tt.start()
            sdev = self._g.storage_device()
            cdev = self._g.compute_device()
            nodes = self.allnodes()
            if sdev.type == 'cpu' and cdev.type == 'cuda' and use_pin:
                pin = self._ctx._get_mem_data_pin(self.layer, len(nodes))
                torch.index_select(self._g.mem.data, 0, nodes, out=pin)
                data = pin.to(cdev, non_blocking=True)
            else:
                data = self._g.mem.data[nodes].to(cdev)
            tt.t_prep_input += tt.elapsed(t_start)
            self._c_mem_data = data

    def _load_mail(self, use_pin=False):
        """Loads the mail to the TGraph's computation device"""
        if self._c_mail is None and self._g.mailbox is not None:
            t_start = tt.start()
            sdev = self._g.storage_device()
            cdev = self._g.compute_device()
            nodes = self.allnodes()
            if sdev.type == 'cpu' and cdev.type == 'cuda' and use_pin:
                pin = self._ctx._get_mail_pin(self.layer, len(nodes))
                torch.index_select(self._g.mailbox.mail, 0, nodes, out=pin)
                data = pin.to(cdev, non_blocking=True)
            else:
                data = self._g.mailbox.mail[nodes].to(cdev)
            tt.t_prep_input += tt.elapsed(t_start)
            self._c_mail = data



import time
from pathlib import Path


class TimeTable(object):
    def __init__(self):
        self.csv = None
        self.reset_epoch()

    def reset_epoch(self):
        """Set all time records to zeros"""
        self.t_epoch = 0.0
        self.t_loop = 0.0
        self.t_eval = 0.0
        self.t_forward = 0.0
        self.t_backward = 0.0
        self.t_sample = 0.0
        self.t_prep_batch = 0.0
        self.t_prep_input = 0.0
        self.t_post_update = 0.0
        self.t_mem_update = 0.0
        self.t_time_zero = 0.0
        self.t_time_nbrs = 0.0
        self.t_self_attn = 0.0

    def start(self):
        # Uncomment for better breakdown timings
        #torch.cuda.synchronize()
        return time.perf_counter()

    def elapsed(self, start):
        # Uncomment for better breakdown timings
        #torch.cuda.synchronize()
        return time.perf_counter() - start

    def print_epoch(self, prefix='  '):
        """Print the timing breakdown of different components in an epoch"""
        lines = f'' \
            f'{prefix}epoch | total:{self.t_epoch:.2f}s loop:{self.t_loop:.2f}s eval:{self.t_eval:.2f}s\n' \
            f'{prefix} loop | forward:{self.t_forward:.2f}s backward:{self.t_backward:.2f}s sample:{self.t_sample:.2f}s prep_batch:{self.t_prep_batch:.2f}s prep_input:{self.t_prep_input:.2f}s post_update:{self.t_post_update:.2f}s\n' \
            f'{prefix} comp | mem_update:{self.t_mem_update:.2f}s time_zero:{self.t_time_zero:.2f}s time_nbrs:{self.t_time_nbrs:.2f}s self_attn:{self.t_self_attn:.2f}s\n'
        print(lines, end='')

    def csv_open(self, path):
        """Close the opened file (if any) and open a new file in write mode"""
        self.csv_close()
        self.csv = Path(path).open('w')

    def csv_close(self):
        """Close the opened file (if any)"""
        if self.csv is not None:
            self.csv.close()
            self.csv = None

    def csv_write_header(self):
        """Write the header line to the CSV file"""
        header = 'epoch,total,loop,eval,' \
            'forward,backward,sample,prep_batch,prep_input,post_update,' \
            'mem_update,time_zero,time_nbrs,self_attn'
        self.csv.write(header + '\n')

    def csv_write_line(self, epoch):
        """Write a line of timing information to the CSV file"""
        line = f'{epoch},{self.t_epoch},{self.t_loop},{self.t_eval},' \
            f'{self.t_forward},{self.t_backward},{self.t_sample},{self.t_prep_batch},{self.t_prep_input},{self.t_post_update},' \
            f'{self.t_mem_update},{self.t_time_zero},{self.t_time_nbrs},{self.t_self_attn}'
        self.csv.write(line + '\n')


# Global for accumulating timings.
tt = TimeTable()



from typing import TYPE_CHECKING, Optional, Tuple
if TYPE_CHECKING:
    from ._graph import TGraph
    from ._context import TContext

import numpy as np
from torch import Tensor

from ._core import TError
from ._block import TBlock
from ._stats import tt


class TBatch(object):
    """
    Represents a batch of temporal edges to process. A thin wrapper with a TGraph reference and without actually
    materializing any arrays until they are needed.
    """

    def __init__(self, g: 'TGraph', range: Tuple[int, int]):
        """
        Internal constructor for creating a TBatch.

        :param TGraph g: The TGraph.
        :param Tuple[int, int] range: The range of edge indices: beginning and ending edge index.
        """
        self._g = g
        self._beg_idx = range[0]
        self._end_idx = range[1]
        self._neg_nodes = None

    def __len__(self) -> int:
        """Returns the total number of edges in the batch."""
        return self._end_idx - self._beg_idx

    @property
    def g(self) -> 'TGraph':
        """Returns the TGraph."""
        return self._g

    @property
    def neg_nodes(self) -> Optional[np.ndarray]:
        """Get the negative nodes."""
        return self._neg_nodes

    @neg_nodes.setter
    def neg_nodes(self, value: np.ndarray):
        """
        Set the negative nodes.

        :param np.ndarray value: An array of negative node samples.
        :raises TError: if value is not a 1-dimensional ndarray.
        """
        if not isinstance(value, np.ndarray):
            raise TError('negative samples must be an ndarray')
        if len(value.shape) != 1:
            raise TError('negative samples must be 1-dimensional')
        self._neg_nodes = value

    def block(self, ctx: 'TContext') -> TBlock:
        """Creates the head TBlock of the batch, including negative nodes if set."""
        t_start = tt.start()
        blk = TBlock(ctx, 0, self.nodes(), self.times())
        tt.t_prep_input += tt.elapsed(t_start)
        return blk

    def block_adj(self, ctx: 'TContext') -> TBlock:
        """Creates the head TBlock with batch edges as neighbors (excluding negative nodes)."""
        dstnodes = self.nodes(include_negs=False)
        srcnodes = self.nodes(include_negs=False, reverse=True)
        dstindex = np.arange(len(dstnodes))
        eids = np.tile(self.eids(), 2)
        ets = self.times(include_negs=False)
        return TBlock(ctx, 0, dstnodes, ets, dstindex, srcnodes, eids, ets)

    def eids(self) -> np.ndarray:
        """
        Returns edge ids of the batch.

        rtype: np.ndarray
        """
        return np.arange(self._beg_idx, self._end_idx, dtype=np.int32)

    def edges(self) -> np.ndarray:
        """
        Returns the edges in the batch as a two-column ndarray, where the first column represents the source
          node index and the second column represents the destination node index.

        rtype: np.ndarray
        """
        return self.g._edges[self._beg_idx:self._end_idx]

    def nodes(self, include_negs=True, reverse=False) -> np.ndarray:
        """
        Returns a node index array: [src, des(, neg)] if reverse is False or [des, src(, src)] if reverse is True.

        :param bool include_negs: Whether to include negative nodes.
        :param bool reverse: Whether to reverse the edges.
        :rtype: np.ndarray
        """
        nids = self.g._edges[self._beg_idx:self._end_idx]
        nids = np.flip(nids, axis=1) if reverse else nids
        nids = nids.T.reshape(-1)
        if self._neg_nodes is not None and include_negs:
            negs = nids[len(self):] if reverse else self._neg_nodes
            nids = np.concatenate([nids, negs])
        return nids.astype(np.int32)

    def times(self, include_dsts=True, include_negs=True) -> np.ndarray:
        """
        Returns timestamps corresponding to the nodes. It retrieves timestamps of the batch edges (as the timestamps for source nodes),
          repeating the timestamps to include destination nodes and negative nodes.

        :param bool include_dst: Whether to include destination nodes of the edges as positive nodes.
        :param bool include_negs: Whether to include negative nodes.
        """
        n_repeats = 2 if include_dsts else 1
        if include_negs and self._neg_nodes is not None:
            n_repeats += 1
        times = self.g._times[self._beg_idx:self._end_idx]
        return np.tile(times, n_repeats).astype(np.float32)

    def split_data(self, data: Tensor) -> Tuple[Tensor, Tensor, Optional[Tensor]]:
        """
        Splits the data into multiple arrays, with each array containing a number of rows equal to the batch size.

        :param Tensor data: The source data to be split.
        :raises TError: If the length of data is not three times the batch size when negative nodes are included or two times otherwise.
        :return: A tuple (src, dst, neg), where neg is None if no negative nodes are specified.
        """
        size = len(self)
        if self._neg_nodes is not None:
            if data.shape[0] != 3 * size:
                raise TError('expected data to have 3 times batch size')
            dst = data[size:2 * size]
            neg = data[2 * size:]
        else:
            if data.shape[0] != 2 * size:
                raise TError('expected data to have 2 times batch size')
            dst = data[size:]
            neg = None
        src = data[:size]
        return (src, dst, neg)


"""
tglite: Temporal GNN Lightweight Framework
"""

from __future__ import annotations

__version__ = "0.0.2"

### Re-exports

from ._graph import TGraph, from_csv
from ._batch import TBatch
from ._block import TBlock
from ._frame import TFrame
from ._memory import Memory
from ._mailbox import Mailbox
from ._sampler import TSampler
from ._context import TContext
from ._core import TError
from . import _utils as utils
from . import nn
from . import op


def iter_edges(g: TGraph, size=1, start=None, end=None) -> EdgesIter:
    """
    Create and return an iterator to generate TBatch

    :param TGraph g: The graph to iterate on.
    :param int size: Number of edges in each mini-batch.
    :rtype: EdgesIter
    :param start: The starting edge index.
    :type start: int or None
    :param end: The ending edge index.
    :type end: int or None
    """
    return EdgesIter(g, size=size, start=start, end=end)


class EdgesIter(object):
    """ An edge iterator of a TGraph."""
    def __init__(self, g: TGraph, size=1, start=None, end=None):
        """
        Create an edge iterator.

        :param TGraph g: The graph it iterates on.
        :param int size: Number of edges in each mini-batch.
        :param start: The starting edge index.
        :type start: int or None
        :param end: The ending edge index.
        :type end: int or None
        """
        self._g = g
        self._size = size
        self._curr = 0 if start is None else start
        self._last = g.num_edges() if end is None else end

    def __iter__(self) -> EdgesIter:
        return self

    def __next__(self) -> TBatch:
        if self._curr < self._last:
            idx = self._curr
            self._curr += self._size
            end = min(self._curr, self._last)
            return TBatch(self._g, range=(idx, end))
        raise StopIteration



from __future__ import annotations

from torch import Tensor

from ._core import TError


class TFrame(dict):
    """A container for storing tensor features."""

    def __init__(self, dim=None):
        """Initialize a TFrame object"""
        super().__init__()
        self._dim = 0 if dim is None else dim

    def dim(self) -> int:
        """Get the leading dimension of stored tensors"""
        return self._dim

    def to(self, device, **kwargs) -> TFrame:
        """Move tensor data in this frame to the specified device.

        :param device: the device to which the tensor data should be moved
        :param **kwargs: additional keyword arguments for the pytorch Tensor.to() method
        :returns: a new TFrame object with the tensor data copied to the specified device
        """
        copy = TFrame(dim=self._dim)
        for key, val in self.items():
            copy[key] = val.to(device, **kwargs)
        return copy

    def __setitem__(self, key, value):
        """
        Set an item in the TFrame

        :raises TError: if the value is not a tensor or if it does not have the expected leading dimension
        """
        if not isinstance(value, Tensor):
            raise TError("expected value to be a tensor")
        if len(value) != self._dim:
            raise TError(f"expected value to have leading dimension of {self._dim}, got {len(value)}")
        super().__setitem__(key, value)



from typing import TYPE_CHECKING, List
if TYPE_CHECKING:
    from ._graph import TGraph

import torch
from torch import Tensor

from ._core import TError


class TContext(object):
    """Graph-level context and scratch space used by the tglite runtime."""

    def __init__(self, g: 'TGraph'):
        """
        Internal constructor for creating a TContext.

        :param TGraph g: The TGraph to operate on.
        """
        self._g = g
        self._training = True

        # pin buffers
        self._efeat_pins = {}
        self._nfeat_pins = {}
        self._mem_data_pins = {}
        self._mail_pins = {}

        # embed caching
        self._cache_enabled = False
        self._cache_dim_emb = None
        self._cache_limit = int(2e6)
        self._cache_tables = {}

        # time precomputation
        self._time_enabled = False
        self._time_window = int(1e4)
        self._time_tables = {}

    @property
    def graph(self) -> 'TGraph':
        """Returns the TGraph it associated with."""
        return self._g

    def train(self):
        """Enables training mode and clear time tables and embedding cache tables."""
        self._training = True
        self._time_tables.clear()
        self._cache_tables.clear()

    def eval(self):
        """Disables training mode."""
        self._training = False

    def need_sampling(self, need: bool):
        """
        Creates tcsr within the TGraph if sampling is needed.

        :param bool need: Whether sampling is required.
        """
        if need:
            self._g._init_tcsr()
        else:
            self._g._tcsr = None

    def enable_embed_caching(self, enabled: bool, dim_embed: int = None):
        """
        Performs embedding cache settings and clear cache tables.

        :param bool enabled: Whether to enable embedding caching.
        :param dim_embed: Dimension of node embeddings.
        :raises TError: If enable is True and dim_embed is None.
        """
        self._cache_enabled = enabled
        if enabled and dim_embed is None:
            raise TError('need dimension of embeddings')
        elif enabled:
            self._cache_dim_emb = dim_embed
        self._cache_tables.clear()

    def set_cache_limit(self, limit: int):
        """
        Sets embedding cache limit and clear cache tables.

        :param int limit: Number of embeddings to cache.
        """
        self._cache_limit = limit
        self._cache_tables.clear()

    def enable_time_precompute(self, enabled: bool):
        """
        Performs time precomputation settings and clear time tables.

        :param bool enabled: Whether to enable embedding caching.
        """
        self._time_enabled = enabled
        self._time_tables.clear()

    def set_time_window(self, window: int):
        """
        Sets length of time window and clear time tables.

        :param int window: Length of time window.
        :raises TError: If int is negative.
        """

        if window < 0:
            raise TError('time window must be non-negative')
        self._time_window = window
        self._time_tables.clear()

    def _get_efeat_pin(self, layer: int, rows: int, dim: int) -> Tensor:
        return self._get_pin(self._efeat_pins, layer, rows, [dim])

    def _get_nfeat_pin(self, layer: int, rows: int, dim: int) -> Tensor:
        return self._get_pin(self._nfeat_pins, layer, rows, [dim])

    def _get_mem_data_pin(self, layer: int, rows: int) -> Tensor:
        return self._get_pin(self._mem_data_pins, layer, rows, [self._g.mem.dim()])

    def _get_mail_pin(self, layer: int, rows: int) -> Tensor:
        return self._get_pin(self._mail_pins, layer, rows, self._g.mailbox.dims())

    def _get_pin(self, cache: dict, layer: int, rows: int, dims: List[int]) -> Tensor:
        """
        Creates/reshapes pinned buffer and returns it.

        :param dict cache: The dictionary of the buffer with layer numbers as keys.
        :param int layer: Which layer pinned data belongs to.
        :param int rows: Number of rows of pinned data.
        :param List[int] dims: Size of each pinned data.
        :return: Pinned buffer.
        :rtype: Tensor
        """
        if layer not in cache:
            shape = tuple([rows] + dims)
            pin = torch.zeros(shape, pin_memory=True)
            cache[layer] = pin
            return pin
        pin = cache[layer]
        if pin.shape[0] < rows or list(pin.shape[1:]) != dims:
            shape = tuple([rows] + dims)
            pin.resize_(shape)
        return pin[:rows]



import os
import numpy as np

from . import _c
from ._core import TError


def get_num_cpus(default=16) -> int:
    cpus = os.cpu_count()
    return default if cpus is None else cpus


def check_edges_times(edges: np.ndarray, times: np.ndarray):
    if edges.shape[0] != times.shape[0]:
        raise TError("edge list and timestamps must have same leading dimension")
    if edges.shape[1] != 2:
        raise TError("edge list must have only 2 columns")
    if edges.dtype != np.int32:
        raise TError("currently only supports int32 node/edge ids")
    if times.dtype != np.float32:
        raise TError("currently only supports float32 timestamps")


def check_num_nodes(edges: np.ndarray, num_nodes: int = None) -> int:
    """Returns the number of nodes in the graph represented by the given edges
    
    :raises TErrror: if the specified number of nodes is less than or equal to the number of distinct nodes present in the edges
    """
    max_nid = int(edges.max())
    num_nodes = max_nid + 1 \
        if num_nodes is None else num_nodes
    if num_nodes <= max_nid:
        raise TError("number of nodes must be greater than max node id")
    return num_nodes


def create_tcsr(edges: np.ndarray, times: np.ndarray, num_nodes: int = None):
    check_edges_times(edges, times)
    num_nodes = check_num_nodes(edges, num_nodes)
    return _c.create_tcsr(edges, times, num_nodes)



from __future__ import annotations
from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from ._block import TBlock
    from ._context import TContext

import torch
import numpy as np
from torch import Tensor

from ._stats import tt
from .op import precomputed_zeros, precomputed_times, edge_reduce, edge_view, edge_softmax


class TimeEncode(torch.nn.Module):

    # A hidden tag used to detect if we have an instance of this builtin encoder
    # so that we can call the more optimized method of generating zeros, without
    # running into situation with circular dependencies.
    __tg_builtin_encoder__ = True

    def __init__(self, dim_time: int):
        '''
        Initializes the TimeEncode module, which encodes time information into a higher-dimensional space.
        
        :param dim_time: dimensionality of the encoded time
        '''
        super().__init__()
        self.w = torch.nn.Linear(1, dim_time)
        self.w.weight = torch.nn.Parameter(torch
            .from_numpy(1 / 10 ** np.linspace(0, 9, dim_time))
            .float().reshape(dim_time, 1))
        self.w.bias = torch.nn.Parameter(torch.zeros(dim_time).float())
        self._z = torch.zeros(1).float()

    def zeros(self, size: int, device):
        '''
        Generates a tensor of zeros with the encoded time dimensionality.
        
        :param size:
        :param device:
        '''
        if self._z.device != torch.device(device):
            self._z = self._z.to(device)
        # expand does not allocate memory
        view = self._z.expand(size)
        return self(view)

    def forward(self, ts: Tensor) -> Tensor:
        '''
        Forward pass of the TimeEncode module. Encodes the input time stamps into a high-dimensional space.
        
        :param ts: input time stamps
        '''
        return torch.cos(self.w(ts.unsqueeze(-1)))


class TemporalAttnLayer(torch.nn.Module):
    def __init__(self, ctx: TContext, num_heads: int,
                 dim_node: int, dim_edge: int, dim_time: int, dim_out: int,
                 dropout=0.1):
        """
        Initializes the Temporal Attention Layer for processing dynamic graphs with temporal features.
        This layer uses multi-head attention mechanism to incorporate node, edge, and time features.

        :param ctx: context object
        :param num_heads: number of heads
        :param dim_node: dimension of node features
        :param dim_edge: dimension of edge features
        :param dim_time: dimension of time features
        :param dim_out: dimension of output features
        :param dropout: dropout rate
        """
        super().__init__()
        assert (dim_out % num_heads == 0)
        self.ctx = ctx
        self.num_heads = num_heads
        self.dim_edge = dim_edge
        self.dim_out = dim_out
        self.time_encode = TimeEncode(dim_time)
        self.w_q = torch.nn.Linear(dim_node + dim_time, dim_out)
        self.w_kv = torch.nn.Linear(dim_node + dim_edge + dim_time, dim_out * 2)
        self.w_out = torch.nn.Linear(dim_node + dim_out, dim_out)
        self.attn_act = torch.nn.LeakyReLU(0.2)
        self.dropout = torch.nn.Dropout(dropout)
        self.layer_norm = torch.nn.LayerNorm(dim_out)

    def forward(self, blk: TBlock) -> Tensor:
        '''
        Forward pass of the Temporal Attention Layer. Applies a time-sensitive attention mechanism over 
        the input graph block (blk) to produce node embeddings. The method handles both cases of graph blocks 
        with and without edges.

        If the block has no edges, a zero-initialized tensor is concatenated with the destination node features.
        For blocks with edges, the method computes attention scores and aggregates neighbor features using 
        the computed attention. 

        :param blk: input graph block
        '''
        if blk.num_edges() == 0:
            dev = blk.dstdata['h'].device
            out = torch.zeros(blk.num_dst(), self.dim_out, dtype=torch.float32, device=dev)
            out = torch.cat([out, blk.dstdata['h']], dim=1)
        else:
            t_start = tt.start()
            zero_time_feat = precomputed_zeros(self.ctx, blk.layer, self.time_encode, blk.num_dst())
            tt.t_time_zero += tt.elapsed(t_start)
            t_start = tt.start()
            nbrs_time_feat = precomputed_times(self.ctx, blk.layer, self.time_encode, blk.time_deltas())
            tt.t_time_nbrs += tt.elapsed(t_start)
            t_start = tt.start()
            Q = torch.cat([blk.dstdata['h'], zero_time_feat], dim=1)
            if self.dim_edge > 0:
                Z = torch.cat([blk.srcdata['h'], blk.efeat(), nbrs_time_feat], dim=1)
            else:
                Z = torch.cat([blk.srcdata['h'], nbrs_time_feat], dim=1)
            del zero_time_feat
            del nbrs_time_feat

            Q = self.w_q(Q)
            Z = self.w_kv(Z)
            K = Z[:, :self.dim_out]
            V = Z[:, self.dim_out:]
            del Z

            Q = edge_view(blk, Q)
            Q = torch.reshape(Q, (Q.shape[0], self.num_heads, -1))
            K = torch.reshape(K, (K.shape[0], self.num_heads, -1))
            V = torch.reshape(V, (V.shape[0], self.num_heads, -1))

            attn = torch.sum(Q * K, dim=2)
            del Q
            del K

            attn = self.attn_act(attn)
            attn = edge_softmax(blk, attn)
            attn = self.dropout(attn)

            out = torch.reshape(V * attn[:, :, None], (V.shape[0], -1))
            del attn

            out = edge_reduce(blk, out, op='sum')
            out = torch.cat([out, blk.dstdata['h']], dim=1)
            tt.t_self_attn += tt.elapsed(t_start)

        out = self.w_out(out)
        out = torch.nn.functional.relu(self.dropout(out))
        out = self.layer_norm(out)
        return out
    



import torch
import numpy as np
from torch import Tensor
from typing import List, Union


class Mailbox(object):
    """A container for node mailbox messages."""

    def __init__(self, num_nodes: int, size: int, dim: int, device=None):
        self._size = size
        self._device = torch.device('cpu' if device is None else device)

        self._mail = torch.zeros((num_nodes, size, dim), device=self._device).squeeze(dim=1)
        self._time = torch.zeros((num_nodes, size), device=self._device).squeeze(dim=1)
        if size > 1:
            self._next = torch.zeros(num_nodes, dtype=torch.long, device=self._device)

    @property
    def mail(self) -> Tensor:
        return self._mail

    @property
    def time(self) -> Tensor:
        return self._time

    @property
    def device(self) -> torch.device:
        return self._device

    def dims(self) -> List[int]:
        return list(self.mail.shape[1:])

    def reset(self):
        self._mail.zero_()
        self._time.zero_()

    def store(self, nids: Union[np.ndarray, Tensor], mail: Tensor, mail_ts: Tensor):
        if not isinstance(nids, Tensor):
            nids = torch.from_numpy(nids).long()
        nids = nids.to(self._device)
        mail = mail.detach().to(self._device)
        mail_ts = mail_ts.detach().to(self._device)
        if self._size == 1:
            self._mail[nids] = mail
            self._time[nids] = mail_ts
        else:
            pos = self._next[nids]
            self._mail[nids, pos] = mail
            self._time[nids, pos] = mail_ts
            self._next[nids] = torch.remainder(pos + 1, self._size)

    def move_to(self, device, **kwargs):
        if device is None or self._device == device:
            return
        self._mail = self._mail.to(device, **kwargs)
        self._time = self._time.to(device, **kwargs)
        self._device = device



import torch
import numpy as np
from torch import Tensor
from pathlib import Path
from typing import Optional, Union

from ._core import TError
from ._frame import TFrame
from ._memory import Memory
from ._mailbox import Mailbox
from ._utils import create_tcsr, check_edges_times, check_num_nodes


class TGraph(object):
    """A container for temporal graph and related tensor data. It initially stores temporal edges in COO format,
    sorted based on timestamp. While performing neighborhood sampling, it uses CSR format for faster lookups. TGLite
    automatically handles the construction and management of these graph formats without intervention from the user."""

    def __init__(self, edges: np.ndarray, times: np.ndarray, num_nodes: int = None):
        """
        Internal constructor for creating a TGraph

        :param np.ndarray edges:
        :param np.ndarray times:
        :param int num_nodes:
        """
        check_edges_times(edges, times)
        self._num_nodes = check_num_nodes(edges, num_nodes)
        self._efeat_frame = TFrame(dim=edges.shape[0])
        self._nfeat_frame = TFrame(dim=self._num_nodes)
        self._edata = TFrame(dim=edges.shape[0])
        self._ndata = TFrame(dim=self._num_nodes)
        self._edges = edges
        self._times = times
        self._tcsr = None
        self._mem = None
        self._mailbox = None
        self._storage_dev = torch.device('cpu')
        self._compute_dev = torch.device('cpu')

    @property
    def efeat(self) -> Optional[Tensor]:
        """Returns edge feature"""
        return self._efeat_frame.get('f')

    @efeat.setter
    def efeat(self, value):
        """
        Sets edge feature

        :param value: edge feature
        """
        if value is None:
            self._efeat_frame.clear()
        else:
            self._efeat_frame['f'] = value

    @property
    def nfeat(self) -> Optional[Tensor]:
        """Returns node feature"""
        return self._nfeat_frame.get('f')

    @nfeat.setter
    def nfeat(self, value):
        """
        Sets node feature

        :param value: edge feature
        """
        if value is None:
            self._nfeat_frame.clear()
        else:
            self._nfeat_frame['f'] = value

    @property
    def edata(self) -> TFrame:
        """Returns edge data"""
        return self._edata

    @property
    def ndata(self) -> TFrame:
        """Returns node data"""
        return self._ndata

    @property
    def mem(self) -> Optional[Memory]:
        """Returns node memory"""
        return self._mem

    @mem.setter
    def mem(self, value: Memory):
        """
        Sets node memory

        :param Memory value: node memory to set
        :raises TError: if value is not a Memory instance or its length doesn't equal to number of nodes,
        or value is not on this TGraph's storage device.
        """
        if not isinstance(value, Memory):
            raise TError('invalid memory object')
        if len(value) != self._num_nodes:
            raise TError('memory number of nodes mismatch')
        if value.device != self._storage_dev:
            raise TError('memory storage device mismatch')
        self._mem = value

    @property
    def mailbox(self) -> Optional[Mailbox]:
        """Returns node mailbox"""
        return self._mailbox

    @mailbox.setter
    def mailbox(self, value: Mailbox):
        """
        Sets mailbox

        :param Mailbox value: mailbox to set
        :raises TError: if value is not a Mailbox instance or its length doesn't equal to number of nodes,
        or value is not on this TGraph's storage device.
        """
        if not isinstance(value, Mailbox):
            raise TError('invalid mailbox object')
        if value.device != self._storage_dev:
            raise TError('mailbox storage device mismatch')
        # ... more checks here ...
        self._mailbox = value

    def storage_device(self) -> torch.device:
        """Returns TGraph's storage device"""
        return self._storage_dev

    def compute_device(self) -> torch.device:
        """Returns TGraph's computing device"""
        return self._compute_dev

    def num_nodes(self) -> int:
        """
        Total number of nodes

        :rtype: int
        """
        return self._num_nodes

    def num_edges(self) -> int:
        """
        Total number of edges

        :rtype: int
        """
        return self._edges.shape[0]

    def set_compute(self, device):
        """Sets computing device"""
        self._compute_dev = torch.device(device)

    def move_data(self, device, **kwargs):
        """Moves tensor data to device while keeping graph on CPU"""
        if self._storage_dev == device:
            return
        self._efeat_frame = self._efeat_frame.to(device, **kwargs)
        self._nfeat_frame = self._nfeat_frame.to(device, **kwargs)
        self._edata = self._edata.to(device, **kwargs),
        self._ndata = self._ndata.to(device, **kwargs),
        if self._mem is not None:
            self._mem.move_to(device, **kwargs)
        if self._mailbox is not None:
            self._mailbox.move_to(device, **kwargs)
        self._storage_dev = device

    def _init_tcsr(self):
        """Creates tcsr of the graph if it doesn't exist"""
        if self._tcsr is None:
            self._tcsr = create_tcsr(self._edges, self._times, num_nodes=self._num_nodes)

    def _get_tcsr(self):
        """Returns the tcsr of the graph"""
        self._init_tcsr()
        return self._tcsr


def from_csv(path: Union[str, Path], skip_first=True) -> TGraph:
    """
    Creates a TGraph from a csv file

    :param path: csv file path
    :type path: str or Path
    :param bool skip_first: whether to skip the first line
    :rtype: TGraph
    :raises TError: if path doesn't exist
    """
    src, dst, ts = [], [], []

    path = Path(path)
    if not path.exists():
        raise TError(f'file does not exist: {path}')

    with path.open() as file:
        if skip_first:
            next(file)
        for line in file:
            line = line.strip().split(',')
            src.append(int(line[0]))
            dst.append(int(line[1]))
            ts.append(float(line[2]))

    src = np.array(src, dtype=np.int32).reshape(-1, 1)
    dst = np.array(dst, dtype=np.int32).reshape(-1, 1)
    edges = np.concatenate([src, dst], axis=1)
    del src
    del dst

    etime = np.array(ts, dtype=np.float32)
    del ts

    return TGraph(edges, etime)



from pathlib import Path
from setuptools import setup
from torch.utils.cpp_extension import BuildExtension, CppExtension


def main():
    curr_dir = Path(__file__).absolute().parent
    setup(
        ext_modules=[
            CppExtension(
                name="tglite._c",
                sources=[
                    "lib/bind.cpp",
                    "lib/cache.cpp",
                    "lib/dedup.cpp",
                    "lib/sampler.cpp",
                    "lib/tcsr.cpp",
                    "lib/utils.cpp"
                ],
                include_dirs=[curr_dir/"include/"],
                extra_compile_args=["-std=c++14", "-fopenmp"],
                extra_link_args=["-fopenmp"]
            )],
        cmdclass={
            "build_ext": BuildExtension
        })


if __name__ == "__main__":
    main()



import torch
import numpy as np
import torch_scatter
from torch import Tensor
from typing import Any, Callable, List, Union

from . import _c
from ._core import TError
from ._block import TBlock
from ._context import TContext
from ._stats import tt


# def find_last_message(uniq_nodes: np.ndarray, sorted_edges: np.ndarray):
#     msg_order, msg_index = _c.find_last_message(uniq_nodes, sorted_edges)
#     msg_order = msg_order.reshape(-1, 2)
#     return msg_order, msg_index


def edge_view(blk: TBlock, data: Tensor) -> Tensor:
    '''
    Reindex the data of edges based on the target node indices in the TBlock.
    
    :param blk:
    :param data:
    '''
    blk._check_has_nbrs()
    assert data.shape[0] == blk._dstdata.dim()
    idx = torch.from_numpy(blk._dstindex)
    idx = idx.to(device=data.device, dtype=torch.long)
    return data[idx]


def edge_softmax(blk: TBlock, data: Tensor) -> Tensor:
    '''
    Computes segmented softmax on given data using edge information from the block.
    
    :param blk:
    :param data:
    '''
    blk._check_has_nbrs()
    size = blk._edata.dim()
    assert data.shape[0] == size
    reindex = torch.from_numpy(blk._dstindex)
    reindex = torch.unique(reindex, return_inverse=True)[1]
    reindex = reindex.to(device=data.device, dtype=torch.long)
    return torch_scatter.scatter_softmax(data, reindex, dim=0, dim_size=size)


def edge_reduce(blk: TBlock, data: Tensor, op='sum') -> Tensor:
    '''
    Computes segmented reduction (e.g. sum or mean) on given data 
    using edge information from the block.
    
    :param blk:
    :param data:
    '''
    blk._check_has_nbrs()
    assert op in ['sum', 'mean'], "currently only supports sum or mean"
    assert data.shape[0] == blk._edata.dim()
    size = blk._dstdata.dim()
    scatter_idx = torch.from_numpy(blk._dstindex)
    scatter_idx = scatter_idx.to(device=data.device, dtype=torch.long)
    return torch_scatter.segment_coo(data, scatter_idx, dim_size=size, reduce=op)


def src_scatter(blk: TBlock, data: Tensor, op='sum') -> Tensor:
    '''
    Aggregates the features of the source node indices in TBlock, 
    using the specified aggregation operation ('sum' or 'mean').
    
    :param blk:
    :param data:
    :param op:
    '''
    blk._check_has_nbrs()
    assert op in ['sum', 'mean'], "currently only supports sum or mean"
    assert data.shape[0] == len(blk._srcnodes)
    uniq_nids, idx = blk.uniq_src()
    idx = idx.to(data.device)
    return torch_scatter.scatter(data, idx, dim=0, dim_size=len(uniq_nids), reduce=op)


def coalesce(blk: TBlock, by='latest') -> TBlock:
    '''
    Segmented operation to reduce source nodes for each destination node 
    by a certain property, such as latest timestamp.
    
    :param blk:
    :param by: method to sample source nodes
    '''
    assert by == 'latest', "currently only supports latest"
    assert blk.has_nbrs() and len(blk.dstnodes) == len(blk.srcnodes)
    uniq_nodes, uniq_idx = np.unique(blk.dstnodes, return_index=True)
    idx = _c.find_latest_uniq(uniq_nodes, blk.dstnodes, blk.ets)
    src = blk.srcnodes[idx]
    eid = blk.eid[idx]
    ets = blk.ets[idx]
    blk._replace_dst(uniq_nodes, blk.dsttimes[uniq_idx])
    blk.set_nbrs(np.arange(len(src)), src, eid, ets)
    return blk


def preload(blk: TBlock, use_pin=True):
    '''
    Prefetch data (e.g. features, memory, mails) needed by the TBlock 
    and its subsequent blocks for computations.

    :param blk:
    :param use_pin: whether to pin memory
    '''
    curr = blk
    while curr.next is not None:
        curr = curr.next
    while curr is not None:
        if curr.num_dst() > 0:
            if curr.next is None:
                curr._load_mail(use_pin=use_pin)
                curr._load_mem_data(use_pin=use_pin)
            if curr.has_nbrs():
                if curr.next is None:
                    curr._load_nfeat(use_pin=use_pin)
                curr._load_efeat(use_pin=use_pin)
        curr = curr.prev


def aggregate(blk: TBlock, fn_or_list: Union[Callable, List[Callable]], key: str = None) -> Any:
    '''
    Performs pull-style multi-hop aggregation from the tail block 
    back towards the given block by applying function to each block, 
    using the key to pass along results.
    
    :param blk:
    :param fn_or_list:
    :param key:
    '''
    while blk.next is not None:
        blk = blk.next
    output = None
    while blk is not None:
        if blk.num_dst() == 0:
            output = blk.run_hooks(output)
        elif isinstance(fn_or_list, List):
            output = blk.apply(fn_or_list[blk.layer])
        else:
            output = blk.apply(fn_or_list)
        t_start = tt.start()
        if blk.prev is not None and output is not None and key:
            if blk._include_prev_dst:
                num_dst = blk.prev.num_dst()
                blk.prev.dstdata[key] = output[:num_dst]
                blk.prev.srcdata[key] = output[num_dst:]
            else:
                blk.prev.srcdata[key] = output
        blk.clear_data()
        blk.clear_hooks()
        tt.t_prep_input += tt.elapsed(t_start)
        blk = blk.prev
    return output


def propagate(blk: TBlock, fn_or_list: Union[Callable, List[Callable]]) -> Any:
    '''
    Performs push-style multi-hop propagation from the given block 
    to the tail block by applying function to each block.
    
    :param blk:
    :param fn_or_list:
    '''
    output = None
    while blk is not None:
        if blk.num_dst() == 0:
            output = blk.run_hooks(output)
        elif isinstance(fn_or_list, List):
            output = blk.apply(fn_or_list[blk.layer])
        else:
            output = blk.apply(fn_or_list)
        blk.clear_data()
        blk.clear_hooks()
        blk = blk.next
    return output


def dedup(blk: TBlock) -> TBlock:
    '''
    Applies the deduplication optimization to the TBlock 
    by rewriting the destination nodes.

    :param blk:
    '''
    if blk.num_dst() == 0:
        return blk
    nodes = blk._dstnodes
    times = blk._dsttimes
    has_dups, nodes, times, inv_idx = _c.dedup_targets(nodes, times)
    if has_dups:
        blk._replace_dst(nodes, times)
        blk.register_hook(_DedupInvertHook(inv_idx))
    return blk


class _DedupInvertHook(object):
    def __init__(self, inv_idx: np.ndarray):
        self.inv_idx = inv_idx

    def __call__(self, _blk: TBlock, output: Tensor) -> Tensor:
        return output[self.inv_idx]


def cache(ctx: TContext, id: int, blk: TBlock, include_first=False):
    '''
    Applies the caching optimization to the TBlock by rewriting 
    the destination nodes and using ctx as scratch space.
    
    :param ctx:
    :param id:
    :param blk:
    :param include_first:
    '''
    if ctx._training or not ctx._cache_enabled:
        return blk
    if blk.prev is None and not include_first:
        # usually not worth it for the head block
        return blk
    if blk.num_dst() == 0:
        return blk

    cache_table = ctx._cache_tables.get(id)
    if cache_table is None:
        cache_table = _c.EmbedTable(ctx._cache_dim_emb, ctx._cache_limit)
        ctx._cache_tables[id] = cache_table

    nodes = blk._dstnodes
    times = blk._dsttimes

    keys = _c.compute_cache_keys(nodes, times)
    hit_idx, embeds = cache_table.lookup(keys, blk.g.compute_device())
    hit_count = torch.sum(hit_idx).item()

    if hit_count == len(nodes):
        blk._replace_dst_empty()
        blk.register_hook(_CacheAllHitsHook(embeds))
    elif hit_count == 0:
        # no need to adjust dst nodes
        blk.register_hook(_CacheAllMissHook(cache_table, keys))
    else:
        miss_idx = (~ hit_idx)
        miss_idx_np = miss_idx.cpu().numpy()
        nodes = nodes[miss_idx_np]
        times = times[miss_idx_np]
        keys = keys[miss_idx_np]
        blk._replace_dst(nodes, times)
        blk.register_hook(_CachePartialHitsHook(cache_table, embeds, miss_idx, keys))

    return blk


class _CacheAllHitsHook(object):
    def __init__(self, embeds: Tensor):
        self.embeds = embeds

    def __call__(self, _blk: TBlock, output: Tensor) -> Tensor:
        '''
        Returns the cached embeddings for all destination nodes.
        
        :param blk:
        :param output:

        '''
        return self.embeds


class _CacheAllMissHook(object):
    def __init__(self, table, miss_keys: np.ndarray):
        self.table = table
        self.miss_keys = miss_keys

    def __call__(self, _blk: TBlock, output: Tensor) -> Tensor:
        '''
        Stores embeddings for all destination nodes.
        
        :param blk:
        :param output:
        '''
        if output.shape[0] != len(self.miss_keys):
            raise TError('dimension mismatch')
        self.table.store(self.miss_keys, output)
        return output


class _CachePartialHitsHook(object):
    def __init__(self, table, embeds: Tensor, miss_idx: Tensor, miss_keys: np.ndarray):
        self.table = table
        self.embeds = embeds
        self.miss_idx = miss_idx
        self.miss_keys = miss_keys

    def __call__(self, _blk: TBlock, output: Tensor) -> Tensor:
        '''
        Returns the cached embeddings for destination nodes with hits,
        and stores the embeddings for destination nodes with misses.
        
        :param blk:
        :param output:
        '''
        if output.shape[0] != len(self.miss_keys):
            raise TError('dimension mismatch')
        self.table.store(self.miss_keys, output)
        self.embeds[self.miss_idx] = output
        return self.embeds


def precomputed_zeros(ctx: TContext, id: int, encoder: Callable, num: int) -> Tensor:
    '''
    Generates a tensor of precomputed zero values encoded by the specified encoder, 
    used for creating a batch of zero time encodings.

    :param ctx:
    :param id:
    :param encoder:
    :param num:
    :return: precomputed zero-initialized tensor of the given size
    '''
    cdev = ctx._g.compute_device()
    if ctx._training or not ctx._time_enabled:
        if getattr(encoder, '__tg_builtin_encoder__', False):
            return encoder.zeros(num, cdev)
        else:
            return encoder(torch.zeros(num, dtype=torch.float, device=cdev))

    time_table = ctx._time_tables.get(id)
    if time_table is None:
        time_table = encoder(torch.arange(
            ctx._time_window + 1, dtype=torch.float, device=cdev))
        ctx._time_tables[id] = time_table

    output = time_table[0].repeat(num, 1)
    output = output.view(num, -1)
    return output


def precomputed_times(ctx: TContext, id: int, encoder: Callable, times: Tensor) -> Tensor:
    '''
    Encodes a tensor of time values using the provided encoder, leverage precomputed values 
    if available in the context's time table. 
    
    :param ctx:
    :param id:
    :param encoder:
    :param times:
    :return: a precomputed tensor of the given times
    '''
    if ctx._training or not ctx._time_enabled:
        return encoder(times)

    time_table = ctx._time_tables.get(id)
    if time_table is None:
        time_table = encoder(torch.arange(
            ctx._time_window + 1, dtype=torch.float, device=ctx._g.compute_device()))
        ctx._time_tables[id] = time_table

    size = times.shape[0]
    hit_count, hit_idx, output, times, inv_idx = \
        _c.find_dedup_time_hits(times, time_table, ctx._time_window)
    uniq_size = times.shape[0]

    if hit_count != uniq_size:
        miss_idx = (~ hit_idx)
        times = times[miss_idx]
        output[miss_idx] = encoder(times.squeeze())

    output = output[inv_idx]
    output = output.view(size, -1)
    return output



from . import _c
from ._core import TError
from ._block import TBlock
from ._utils import get_num_cpus
from ._stats import tt


class TSampler(object):

    def __init__(self, num_nbrs: int, strategy='recent', num_threads: int = None):
        """
        Internal constructor for creating a TSampler

        :param int num_nbrs: number of neighbors
        :param str strategy: sampling strategy, 'recent' or 'uniform'
        :param int num_threads: number of threads for parallel sampling, set to number of cpus if not provided
        :raises TError: if strategy is not in ['recent', 'uniform']
        """

        if strategy not in ['recent', 'uniform']:
            raise TError(f'sampling strategy not supported: {strategy}')

        self._n_nbrs = num_nbrs
        self._strategy = strategy
        self._n_threads = get_num_cpus() \
            if num_threads is None else num_threads

        self._sampler = _c.TemporalSampler(
            self._n_threads,
            self._n_nbrs,
            self._strategy == 'recent')

    def sample(self, blk: TBlock) -> TBlock:
        """Updates block with sampled 1-hop source neighbors
        
        :returns: updated block
        """
        t_start = tt.start()
        if blk.num_dst() > 0:
            block = self._sampler.sample(blk._g._get_tcsr(), blk._dstnodes, blk._dsttimes)
            blk.set_nbrs(
                block.copy_dstindex(),
                block.copy_srcnodes(),
                block.copy_eid(),
                block.copy_ets())
        tt.t_sample += tt.elapsed(t_start)
        return blk
